#include "FunctionsManager.h"

namespace FivePointAmplitudes {

namespace detail {
template <typename T> std::complex<double> to_double(const std::complex<T>& c) { return {to_double(c.real()), to_double(c.imag())}; }
} // namespace detail

size_t FunctionsManager::add(const FunctionID& fid) {
    static std::mutex m;
    std::lock_guard<std::mutex> lock(m);

    if (locked) { throw std::logic_error("FunctionsManager does not accept new functions after the first evaluation!"); }

    auto find = _function_indices.find(fid);

    if (find != _function_indices.end()) return find->second;

    _evaluators.push_back(fid.get_evaluator<R>());
    last_values.push_back({});

#ifdef FivePointAmplitudes_QD_ENABLED
    _evaluators_HP.push_back(fid.get_evaluator<dd_real>());
    last_values_HP.push_back({});
#endif // FivePointAmplitudes_QD_ENABLED


    size_t new_id = _evaluators.size() - 1;

    _function_indices[fid] = new_id;

    return new_id;
}

const std::vector<FunctionsManager::C>& FunctionsManager::eval(const Kin<R>& kin) {
    static std::mutex m;
    std::lock_guard<std::mutex> lock(m);

    const size_t N = _evaluators.size();

    assert(last_values.size() == N);

    if (kin.v == last_kin.v) return last_values;

    for (size_t i = 0; i < N; ++i) { last_values[i] = _evaluators[i](kin); }
    last_kin = kin;

    locked = true;

    return last_values;
}

#ifdef FivePointAmplitudes_QD_ENABLED 
const std::vector<FunctionsManager::CHP>& FunctionsManager::eval(const Kin<RHP>& kin) {return eval_HP(kin);}

const std::vector<FunctionsManager::CHP>& FunctionsManager::eval_HP(const Kin<RHP>& kin) {
    static std::mutex m;
    std::lock_guard<std::mutex> lock(m);

    const size_t N = _evaluators.size();

    assert(last_values_HP.size() == N);

    if (kin.v == last_kin_HP.v) return last_values_HP;

    for (size_t i = 0; i < N; ++i) {
        last_values_HP[i] = _evaluators_HP[i](kin); 
        last_values[i] = detail::to_double(last_values_HP[i]); 
    }
    last_kin_HP = kin;

    for (size_t i = 0; i < kin.v.size(); ++i) {
        last_kin.v[i] = to_double(kin.v[i]);
    }
    last_kin = Kin<R>{last_kin.v};

    locked = true;

    return last_values_HP;
}
#endif // FivePointAmplitudes_QD_ENABLED

f_type from_string(std::string s) {
    if ( s == "F" ) return f_type::f;
    if ( s == "bc" ) return f_type::bc;
    if ( s == "bci" ) return f_type::bci;
    if ( s == "str5" ) return f_type::str5;
    if ( s == "one" ) return f_type::one;

    throw std::runtime_error("Unidentified f_type "+s);
}

    
} // FivePointAmplitudes
